package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.core.cache.ParamCache;
import cn.com.qmth.stmms.ms.core.domain.Level;
import cn.com.qmth.stmms.ms.core.domain.task.MarkTask;
import cn.com.qmth.stmms.ms.core.domain.user.MarkUser;
import cn.com.qmth.stmms.ms.core.repository.MarkUserRepo;
import cn.com.qmth.stmms.ms.marking.service.arbitration.ArbitrationService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;
import java.util.stream.Stream;

/**
 * 定档
 * Created by zhengmin on 2016/9/22.
 */
@Service
public class DetermineLevelService {

    public enum DeterType {
        /**
         * 多数
         */
        MAJORITY,
        /**
         * 权重
         */
        WEIGHT,
        /**
         * 仲裁
         */
        ARBITRATE,
        /**
         * 打回
         */
        REJECT
    }

    @Autowired
    private MarkUserRepo markUserRepo;

    /**
     * 计算定档值（规则：1.不是过半定档，取权重，2.是过半定档，没有取优，取过半定档，3.是过半定档，是取优，取过半定档和权重最优值）
     *
     * @param workId 工作id
     * @param levels 档位集合
     * @param markTasks    任务集合
     */
    public DeterResult determine(Long workId, List<Level> levels, List<MarkTask> markTasks) {
        //定档
        boolean majority = Optional.ofNullable(ParamCache.paramMap.get(workId).getMajority()).orElse(1) == 1;
        DeterResult majorityResult = null;
        DeterResult weightResult = null;
        if (majority) {
            MarkTask[] tasks = markTasks.toArray(new MarkTask[0]);
            majorityResult = calcMajority(tasks);
        }
        if (majorityResult != null) {
            // 是否取优（过半定档和权重，取最优值）
            boolean takeBest = Optional.ofNullable(ParamCache.paramMap.get(workId).getTakeBest()).orElse(1) == 1;
            if (takeBest) {
                Map<String, Integer> levelMap = levels.stream().collect(Collectors.toMap(Level::getCode, Level::getLevelValue));

                weightResult = calcWeight(levels, markTasks);
                // 取优
                int majorityLevelValue = levelMap.get(majorityResult.getResult());
                int weightLevelValue = levelMap.get(weightResult.getResult());
                // 值越小，档位越大
                if (majorityLevelValue > weightLevelValue) {
                    return weightResult;
                } else {
                    return majorityResult;
                }
            } else {
                return majorityResult;
            }
        }
        // 无过半档位或者未开启过半定档
        else {
            // 仲裁档位差
            int deviation = Optional.ofNullable(ParamCache.paramMap.get(workId).getDeviation()).orElse(4);
            // 超过仲裁档位差
            Set<Integer> sources = markTasks.stream().map(MarkTask::getLevelValue).collect(Collectors.toSet());
            if (ArbitrationService.overDeviation(deviation, sources)) {
                boolean autoCallBack = Optional.ofNullable(ParamCache.paramMap.get(workId).getAutoCallback()).orElse(0) == 1;
                // 开启自动打回
                if (autoCallBack) {
                    int cumulateError = Optional.ofNullable(ParamCache.paramMap.get(workId).getCumulativeError()).orElse(9);
                    List<String> markerIds = ArbitrationService.calcCumulate(cumulateError, markTasks);
                    // 不是所有评卷员累计误差都大于累计误差值
                    if (!markerIds.isEmpty() && markerIds.size() != markTasks.size()) {
                        return new DeterResult(null, DeterType.REJECT, markerIds);
                    }
                }
                // 打回科组长
                return new DeterResult(null, DeterType.ARBITRATE);
            } else {
                return calcWeight(levels, markTasks);
            }
        }
    }

    /**
     * 计算过半定档值
     *
     * @param tasks 任务集合
     */
    private DeterResult calcMajority(MarkTask[] tasks) {
        //如果相同的数量大于半数，直接定档
        Stream<MarkTask> taskStream = Stream.of(tasks);
        Optional<Map.Entry<String, Long>> optional = taskStream
                .collect(Collectors.groupingBy(MarkTask::getResult, Collectors.counting()))
                .entrySet()
                .stream()
                .max(Map.Entry.comparingByValue());
        if (optional.isPresent()) {
            Map.Entry<String, Long> entry = optional.get();
            if (entry.getValue() > tasks.length / 2) {
                return new DeterResult(entry.getKey(), DeterType.MAJORITY);
            }
        }
        return null;
    }

    /**
     * 计算权重值
     *
     * @param markTasks 任务集合
     */
    private DeterResult calcWeight(List<Level> levels, List<MarkTask> markTasks) {
        MarkTask[] tasks = markTasks.toArray(new MarkTask[0]);
        //通过权重计算，平均值四舍五入
        double[] values = new double[tasks.length];
        double[] weights = new double[tasks.length];
        for (int i = 0; i < tasks.length; i++) {
            String levelResult = tasks[i].getResult();
            Level level = levels.stream().filter(m -> m.getCode().equals(levelResult)).findFirst().orElse(null);
            if (level == null) {
                throw new RuntimeException("无此档位");
            }
            MarkUser marker = markUserRepo.findOne(tasks[i].getMarkerId());
            double markerWeight = marker.getWeight();
            values[i] = level.getWeight() * markerWeight;
            weights[i] = markerWeight;
        }
        int avg = (int) Math.round(DoubleStream.of(values).sum() / DoubleStream.of(weights).sum());
        Optional<Level> levelOptional = levels.stream().filter(o -> o.getMaxScore() >= avg && o.getMinScore() <= avg).findFirst();
        Level level = levelOptional.orElse(null);
        if (level == null) {
            levels = levels.stream().sorted(Comparator.comparing(Level::getMaxScore).reversed()).collect(Collectors.toList());
            if (avg > levels.get(0).getMaxScore()) {
                level = levels.get(0);
            }
            levels = levels.stream().sorted(Comparator.comparing(Level::getMinScore)).collect(Collectors.toList());
            if (avg < levels.get(0).getMinScore()) {
                level = levels.get(0);
            }
        }
        return new DeterResult(level.getCode(), DeterType.WEIGHT);
    }
}

class DeterResult {
    private String result;
    private DetermineLevelService.DeterType deterType;
    private List<String> rejectTasks;

    public DeterResult(String result, DetermineLevelService.DeterType deterType) {
        this.result = result;
        this.deterType = deterType;
    }

    public DeterResult(String result, DetermineLevelService.DeterType deterType, List<String> rejectTasks) {
        this.result = result;
        this.deterType = deterType;
        this.rejectTasks = rejectTasks;
    }

    public String getResult() {
        return result;
    }

    public void setResult(String result) {
        this.result = result;
    }

    public DetermineLevelService.DeterType getDeterType() {
        return deterType;
    }

    public void setDeterType(DetermineLevelService.DeterType deterType) {
        this.deterType = deterType;
    }

    public List<String> getRejectTasks() {
        return rejectTasks;
    }

    public void setRejectTasks(List<String> rejectTasks) {
        this.rejectTasks = rejectTasks;
    }
}